package tagInducer;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import utils.FileUtils;
import utils.MapUtil;
import utils.StringUtils;

public class Corpus {

	/** The full corpus (arrays of sentences) */
	private int[][] corpus;

	/** The full corpus tags (arrays of sentences) */
	private int[][] corpusGoldTags;

	/** Map from integers to word types */
	private Map<Integer, String> int2WordTypes = new HashMap<Integer, String>();
	/** Map from word types to integers */
	private Map<String, Integer> wordTypes2Int = new HashMap<String, Integer>();

	/** 1D array of the corpus (bag-of-words) */
	private int[] words;

	/** Map from integers to gold-standard tags  */
	private Map<Integer, String> int2GoldTags = new HashMap<Integer, String>();
	/** Map from gold-standard tags to integers */
	private Map<String, Integer> goldTags2Int = new HashMap<String, Integer>();

	/** 1D array of the tags of the corpus (for evaluation) */
	private int[] goldTags;

	/** A map of feature vectors per word type.<br>
	 * <code>String</code> will contain the feature type (left, right, ...)
	 * and <code>int[][]</code> is a MxF array of features per type.<br> 
	 */
	private Map<String, int[][]> featureVectors = new HashMap<String, int[][]>();

	private FileUtils f = new FileUtils();

	/** Number of feature (most frequent) words */
	private int NUM_CONTEXT_FEAT_WORDS = 100;

	/** Map for word frequencies (wordType -> count) */
	private Map<Integer, Integer> wordFreq;

	/** Whether the corpus has gold-standard tag annotation */
	private boolean HAS_TAGS = true;

	public boolean hasTags() {
		return HAS_TAGS;
	}

	private Map<Integer, Integer> contextWordMap = new HashMap<Integer, Integer>();

	List<Integer> contextWordList;

	private String morphMethod, morphFile;

	private StringUtils s = new StringUtils();

	/**
	 * Extracts a corpus from input
	 * @param input The file name
	 */
	public Corpus(File inputFile, int numContextFeats){
		if (numContextFeats != 0) NUM_CONTEXT_FEAT_WORDS = numContextFeats;

		//Read the corpus
		readCorpus(inputFile);

		//Populate words
		int wordInd = 0;
		for (int sentence = 0; sentence < corpus.length; sentence++){
			for (int word = 0; word < corpus[sentence].length; word++){
				words[wordInd] = corpus[sentence][word];
				wordInd++;
			}
		}

		//Create a list of feature words (N most frequent words)
		wordFreq = new HashMap<Integer, Integer>();
		//Get the word counts
		for (int word:words){
			if (wordFreq.containsKey(word)) wordFreq.put(word, wordFreq.get(word)+1);
			else wordFreq.put(word, 1);
		}
		//Resort wrt frequency and prune
		contextWordList = MapUtil.sortByValueList(wordFreq).subList(0, NUM_CONTEXT_FEAT_WORDS);
		wordInd = 0;

		wordInd = 0;
		for (int word:contextWordList){
			contextWordMap.put(word, wordInd);
			wordInd++;
		}	
	}

	public void addAllContextFeats() throws IOException{
		//Add the left and right NULL features (+2 in total)
		int[][] features = new int[wordTypes2Int.size()][(NUM_CONTEXT_FEAT_WORDS*2)+2];
		//This fills the feature vectors up to 2*numContextWords
		int[][] leftFeatures = new int[wordTypes2Int.size()][NUM_CONTEXT_FEAT_WORDS+1];
		for (int[] sentence: corpus) {
			//For every target word in the sentence
			for (int pos=0; pos < sentence.length; pos++) {
				int wordIndex = sentence[pos];
				int prevContextPosition = pos-1;
				//Add the previous context word
				if (prevContextPosition >= 0) {
					int contextIndex =  sentence[prevContextPosition];
					if (contextWordMap.containsKey(contextIndex))
						leftFeatures[wordIndex][contextWordMap.get(contextIndex)]++;
					//Null left context
					else leftFeatures[wordIndex][NUM_CONTEXT_FEAT_WORDS]++;
				}
			}
		}
		int[][] rightFeatures = new int[wordTypes2Int.size()][NUM_CONTEXT_FEAT_WORDS+1];
		for (int[] sentence: corpus) {
			//For every target word in the sentence
			for (int pos=0; pos < sentence.length; pos++) {
				int wordIndex = sentence[pos];
				int nextContextPosition = pos+1;
				//Add the next context word
				if (nextContextPosition < sentence.length) {
					int contextIndex =  sentence[nextContextPosition];
					if (contextWordMap.containsKey(contextIndex))
						rightFeatures[wordIndex][contextWordMap.get(contextIndex)]++;
					//Null right context
					else rightFeatures[wordIndex][NUM_CONTEXT_FEAT_WORDS]++;
				}
			}
		}
		for (int i=0; i<leftFeatures.length; i++)
			System.arraycopy(leftFeatures[i], 0, features[i], 0, leftFeatures[i].length);
		for (int i=0; i<rightFeatures.length; i++)
			System.arraycopy(rightFeatures[i], 0, features[i], leftFeatures[i].length, rightFeatures[i].length);
		featureVectors.put("ALL_CONTEXT", features);
	}

	public void addLeftContextFeats() throws IOException{
		int[][] features = new int[wordTypes2Int.size()][NUM_CONTEXT_FEAT_WORDS+1];
		for (int[] sentence: corpus) {
			//For every target word in the sentence
			for (int pos=0; pos < sentence.length; pos++) {
				int wordIndex = sentence[pos];
				int prevContextPosition = pos-1;
				//Add the previous context word
				if (prevContextPosition >= 0) {
					int contextIndex =  sentence[prevContextPosition];
					if (contextWordMap.containsKey(contextIndex))
						features[wordIndex][contextWordMap.get(contextIndex)]++;
					//Null right context
					else features[wordIndex][NUM_CONTEXT_FEAT_WORDS]++;
				}
			}
		}
		featureVectors.put("LEFT", features);
	}

	public void addRightContextFeats() throws IOException{
		int[][] features = new int[wordTypes2Int.size()][NUM_CONTEXT_FEAT_WORDS+1];
		for (int[] sentence: corpus) {
			//For every target word in the sentence
			for (int pos=0; pos < sentence.length; pos++) {
				int wordIndex = sentence[pos];
				int nextContextPosition = pos+1;
				//Add the previous context word
				if (nextContextPosition < sentence.length) {
					int contextIndex =  sentence[nextContextPosition];
					if (contextWordMap.containsKey(contextIndex))
						features[wordIndex][contextWordMap.get(contextIndex)]++;
					//Null right context
					else features[wordIndex][NUM_CONTEXT_FEAT_WORDS]++;
				}
			}
		}
		featureVectors.put("RIGHT", features);
	}

	public void extendMorph(boolean extendedMorphFeats) throws IOException{
		Map<String, Map<String, Integer>> morphFeatCounts = readMorph();

		//Add the NULL morph feature
		int numMorphFeats = morphFeatCounts.size()+1;
		int[][] morphFeats = new int[wordTypes2Int.size()][numMorphFeats];
		for (int[] feats:morphFeats) Arrays.fill(feats, 0);

		//Now fill the last part of the feature vectors with morphological features
		//Get the word index
		int wordIndex; int featIndex = 0;
		for (String morphFeat:morphFeatCounts.keySet()){
			for (String word:morphFeatCounts.get(morphFeat).keySet()){
				if (!wordTypes2Int.containsKey(word)) continue;
				wordIndex = wordTypes2Int.get(word);
				morphFeats[wordIndex][featIndex]++;
			}
			featIndex++;
		}
		//Add the NULL feature counts
		for (int i = 0; i < morphFeats.length; i++){
			if (isNull(morphFeats[i])) morphFeats[i][numMorphFeats-1]++;
		}
		
		if (extendedMorphFeats){
			//Indices:
			//0: hasInitialCapital
			//1: hasDigit
			//2: hasHyphen
			//3: hasPunctuation
			int[][] extendedFeats = new int[wordTypes2Int.size()][4];
			for (int[] feats:extendedFeats) Arrays.fill(feats, 0);
			
			for (int wordInd:int2WordTypes.keySet()){
				String word = int2WordTypes.get(wordInd);
				if (word.matches("^[A-Z].*"))extendedFeats[wordInd][0]++;
				if (word.matches(".*[0-9].*")) extendedFeats[wordInd][1]++;
				if (word.contains("-")) extendedFeats[wordInd][2]++;
				if (word.matches(".*\\W.*"))extendedFeats[wordInd][3]++;
			}
			int[][] newFeats = new int[wordTypes2Int.size()][numMorphFeats+4];
			for (int i=0; i<wordTypes2Int.size(); i++){
				System.arraycopy(morphFeats[i], 0, newFeats[i], 0, morphFeats[i].length);
				
				System.arraycopy(extendedFeats[i], 0, newFeats[i], 
						morphFeats[i].length, extendedFeats[i].length);
			}
			featureVectors.put("MORPH", newFeats);
		}
		else featureVectors.put("MORPH", morphFeats);
	}

	private boolean isNull(int[] vector) {
		for (int val:vector){
			if (val != 0) return false;
		}
		return true;
	}

	private Map<String, Map<String, Integer>> readMorph() throws IOException{
		Map<String, Map<String, Integer>> morphFeatCounts = new HashMap<String, Map<String, Integer>>();
		if (morphMethod.equalsIgnoreCase("morfessor")){
			BufferedReader in = f.createIn(morphFile);
			String line, word, suffix = null;
			String[] splits;			
			while ((line = in.readLine())!=null){
				if (line.startsWith("#")) continue;
				splits = line.split("\\s+");
				word = "";
				if (splits.length < 2) continue;
				//Ignore the number for now
				for (int i = 1; i < splits.length; i++){
					if (splits[i].startsWith("+")) continue;
					word += splits[i].substring(0, splits[i].indexOf("/"));
				}
				if (splits[splits.length-1].contains("SUF")){
					suffix = s.joinSuffixes(splits);
					if (morphFeatCounts.containsKey(suffix)){
						morphFeatCounts.get(suffix).put(word, 1);
					}
					else {
						Map<String, Integer> temp = new HashMap<String, Integer>();
						temp.put(word, 1);
						morphFeatCounts.put(suffix, temp);
					}
				}
			}
		}
		//Use the final 2 letters of each word as features
		else {
			for (int wordInd:words){
				String wordType = int2WordTypes.get(wordInd);
				if (!s.checkString(wordType, 4)) continue;
				String suffix = wordType.substring(wordType.length()-2);
				if (morphFeatCounts.containsKey(suffix)) {
					if (!morphFeatCounts.get(suffix).containsKey(wordType))
						morphFeatCounts.get(suffix).put(wordType, 1);
				}
				else {
					Map<String, Integer> temp = new HashMap<String, Integer>();
					temp.put(wordType, 1);
					morphFeatCounts.put(suffix, temp);
				}
			}
		}
		return morphFeatCounts;
	}

	public int[] initFreq(int numClusters){
		int[] assignments = new int[wordTypes2Int.size()];
		List<Integer> skip = new ArrayList<Integer>();
		
		for (int i=0; i<numClusters; i++){
			Object word = contextWordList.get(i);
			int wordIndex = Integer.parseInt(word.toString());
			assignments[wordIndex] = i;
			skip.add(wordIndex);
		}
		
		for (int type = 0; type<wordTypes2Int.size(); type++){
			if (skip.contains(type)) continue;
			int cluster = (int) (Math.random() * numClusters);
			assignments[type] = cluster;
		}
		return assignments;
	}

	public void extendAligns(String alignsFile, String corpusFile, String[] langs) throws IOException{
		Map<String, int[][]> alignFeatures = new HashMap<String, int[][]>();
		Map<String, int[][]> corpora = new HashMap<String, int[][]>();
		Map<String, Map<Integer, Integer>> corporaContextWordMaps = new HashMap<String, Map<Integer, Integer>>();
		int[] contextWords = new int[100];
		//XXX Assume we're using multext files
		String corpusLang = corpusFile.substring(corpusFile.lastIndexOf('_')+1,corpusFile.lastIndexOf('.'));
		for (String lang:langs){
			int[][] temp = readLangCorpus("data/multext_"+lang+".tagged");
			corpora.put(lang, temp);
			contextWords = new int[contextWords.length];
			Map<Integer, Integer> wordFreq = new HashMap<Integer, Integer>();
			//Get the word counts
			for (int[] sentence:temp){
				for (int word:sentence){
					if (wordFreq.containsKey(word)) wordFreq.put(word, wordFreq.get(word)+1);
					else wordFreq.put(word, 1);
				}
			}
			//Resort wrt frequency and prune
			List<Integer> contextWordList= MapUtil.sortByValueList(wordFreq).subList(0, contextWords.length);
			Map<Integer, Integer> contextWordMap = new HashMap<Integer, Integer>();
			int wordInd = 0;
			for (int word:contextWordList){
				contextWordMap.put(word, wordInd);
				wordInd++;
			}
			int[][] features = new int[wordTypes2Int.size()][contextWords.length];
			alignFeatures.put(lang, features);
			corporaContextWordMaps.put(lang, contextWordMap);
		}

		//Assume lang1.sentNo|wordNo-lang2.sentNo|wordNo
		BufferedReader in = f.createIn(alignsFile);
		String line, lang;
		int srcWordTypeInd, wordP, wordN;
		String[] splits;
		BigWhile: while ((line = in.readLine()) != null){
			if (!line.contains(corpusLang)) continue;
			for (String tempLang:langs) {
				if (!line.contains(tempLang)) continue BigWhile;
			}
			wordP = -1; wordN = -1;
			splits = line.split("\\W");
			if (splits[0].equals(corpusLang)){
				srcWordTypeInd = corpus[Integer.parseInt(splits[1])][Integer.parseInt(splits[2])];
				lang = splits[3];
				//Catch some nasty alignment errors
				if (corpora.get(lang)[Integer.parseInt(splits[4])].length <= Integer.parseInt(splits[5])) continue;
				if (Integer.parseInt(splits[5])-1 > 0)
					wordP = corpora.get(lang)[Integer.parseInt(splits[4])][Integer.parseInt(splits[5])-1];
				if (corpora.get(lang)[Integer.parseInt(splits[4])].length <Integer.parseInt(splits[5])+1)
					wordN = corpora.get(lang)[Integer.parseInt(splits[4])][Integer.parseInt(splits[5])+1];
			}
			else {
				srcWordTypeInd = corpus[Integer.parseInt(splits[4])][Integer.parseInt(splits[5])];
				lang = splits[0];
				//Catch some nasty alignment errors
				if (corpora.get(lang)[Integer.parseInt(splits[1])].length <= Integer.parseInt(splits[2])) continue;
				if (Integer.parseInt(splits[2])-1 > 0)
					wordP = corpora.get(lang)[Integer.parseInt(splits[1])][Integer.parseInt(splits[2])-1];
				if (corpora.get(lang)[Integer.parseInt(splits[1])].length <	Integer.parseInt(splits[2])+1)
					wordN = corpora.get(lang)[Integer.parseInt(splits[1])][Integer.parseInt(splits[2])+1];
			}
			if (corporaContextWordMaps.get(lang).containsKey(wordP) && wordP!=-1) alignFeatures.get(lang)[srcWordTypeInd][corporaContextWordMaps.get(lang).get(wordP)]++;
			if (corporaContextWordMaps.get(lang).containsKey(wordN) && wordN!=-1) alignFeatures.get(lang)[srcWordTypeInd][corporaContextWordMaps.get(lang).get(wordN)]++;
		}

		//Account for one NULL feature for all languages
		int numTotFeats = contextWords.length*alignFeatures.size();
		int features[][] = new int[wordTypes2Int.size()][numTotFeats+1];
		int curPos = 0;
		int[] nullFeats = new int[wordTypes2Int.size()];
		for (String language:corporaContextWordMaps.keySet()){
			int numAlignFeats = corporaContextWordMaps.get(language).size();
			int temp[][] = new int[wordTypes2Int.size()][numAlignFeats+1];
			for (int wordType = 0; wordType < wordTypes2Int.size(); wordType++) {
				for (int feat = 0; feat < numAlignFeats; feat++){
					temp[wordType][feat] = alignFeatures.get(language)[wordType][feat];
				}
				//Cumulate NULL features for each language (they are emitted every time no alignment feature is generated)
				//Effectively we take the number of tokens (sum of context features) and subtract the number of algin features emitted
				int tokensPerType = sumContextFeats(wordType);
				int alignTokens = sumAlignFeats(temp[wordType]);
				assert (tokensPerType >= alignTokens);
				nullFeats[wordType] += tokensPerType-alignTokens;
			}
			for (int i = 0; i < temp.length; i++) System.arraycopy(temp[i], 0, features[i], curPos, temp[i].length);
			curPos += temp[0].length;
		}
		//Add the total NULL feature emissions
		for (int wordType = 0; wordType < wordTypes2Int.size(); wordType++)
			features[wordType][numTotFeats] = nullFeats[wordType];
		featureVectors.put("ALIGNS", features);
	}


	private int sumAlignFeats(int[] vector) {
		int sum = 0;
		for (int i = 0; i < vector.length; i++) sum += vector[i];
		return sum;
	}

	private int sumContextFeats(int type) {
		int sum = 0;
		if (featureVectors.containsKey("ALL_CONTEXT")) {
			int allSumL = 0; int allSumR = 0;
			int size = featureVectors.get("ALL_CONTEXT")[type].length;
			//Use half of the features to get the number of tokens
			//Check both halves to account for sentence-boundary tokens
			for (int i = 0; i < (size)/2; i++) 
				allSumL += featureVectors.get("ALL_CONTEXT")[type][i];
			for (int i = (size)/2; i < size; i++) 
				allSumR += featureVectors.get("ALL_CONTEXT")[type][i];
			if (allSumL > allSumR) sum = allSumL;
			else sum = allSumR;
		}
		else {
			for (int i = 0; i < featureVectors.get("LEFT")[type].length; i++) 
				sum += featureVectors.get("LEFT")[type][i];
		}
		return sum;
	}

	public int[][] readLangCorpus(String file) throws IOException{
		String line;
		int wordTypeInd = 0, sentenceInd = 0;
		Map<String, Integer> wordTypes2Int = new HashMap<String, Integer>();

		int[][] corpus = new int[s.count(file)][];

		//Read the corpus
		BufferedReader in = f.createIn(file);
		while ((line = in.readLine())!=null){
			//Map word types to integers
			int[] lineWords = new int[line.split("\\s+").length];
			int wordInd = 0;
			for (String word:line.split("\\s+")){
				word = s.extractWord(word);
				//Get the word index
				if (!wordTypes2Int.containsKey(word)) {
					wordTypes2Int.put(word, wordTypeInd);
					wordTypeInd++;
				}
				lineWords[wordInd] = wordTypes2Int.get(word);
				wordInd++;
			}
			corpus[sentenceInd] = lineWords;
			sentenceInd++;
		}
		in.close();

		return corpus;
	}

	private void readCorpus(File file){
		String line;
		int wordTypeInd = 0, tagInd = 0, sentenceInd = 0, totalWords = 0;

		//Get the corpus statistics (and also check for tags)
		try {
			BufferedReader in = f.createIn(file);
			while ((line = in.readLine())!=null){
				sentenceInd++;
				totalWords+=line.split("\\s+").length;
				if (s.checkForTags(line) && HAS_TAGS) HAS_TAGS = true;
				else HAS_TAGS = false;
			}
			in.close();
		} 
		catch (IOException e) {
			System.err.println("Error reading the corpus file.");
			System.exit(-1);
		}

		corpus = new int[sentenceInd][];
		words = new int[totalWords];
		if (HAS_TAGS) {
			corpusGoldTags = new int[sentenceInd][];
			goldTags = new int[totalWords];
		}
		sentenceInd = 0;

		//Read the corpus
		try {
			BufferedReader in = f.createIn(file);
			while ((line = in.readLine())!=null){
				//Map word types to integers
				int[] lineWords = new int[line.split("\\s+").length];
				//Map tags to integers
				int[] lineTags = new int[line.split("\\s+").length];
				int wordInd = 0;
				for (String word:line.split("\\s+")){
					String tag = null;
					//If tags exist, first extract them...
					if (HAS_TAGS) {
						tag = s.extractTag(word);
						word = s.extractWord(word);

						//... and then get the tag index
						if (!goldTags2Int.containsKey(tag)){
							goldTags2Int.put(tag, tagInd);
							int2GoldTags.put(tagInd, tag);
							tagInd++;
						}
						lineTags[wordInd] = goldTags2Int.get(tag);
					}

					//Get the word index
					if (!wordTypes2Int.containsKey(word)) {
						wordTypes2Int.put(word, wordTypeInd);
						int2WordTypes.put(wordTypeInd, word);
						wordTypeInd++;
					}
					lineWords[wordInd] = wordTypes2Int.get(word);
					wordInd++;
					totalWords++;
				}
				corpus[sentenceInd] = lineWords;
				if (HAS_TAGS) corpusGoldTags[sentenceInd] = lineTags;
				sentenceInd++;
			}
			in.close();
		} 
		catch (IOException e) {
			System.err.println("Error reading the corpus file.");
			System.exit(-1);
		}
	}

	/**
	 * Outputs the tagged corpus
	 * @param classes The list of induced tagged
	 * @param outFile The name of the output file
	 */
	public void writeTagged(int[] classes, String outFile) throws IOException{
		String outLine = "";
		int wordInd = 0;
		BufferedWriter out = f.createOut(outFile);
		for (int sentence = 0; sentence < corpus.length; sentence++){
			for (int word = 0; word < corpus[sentence].length; word++){
				outLine += int2Word(words[wordInd]) + "/" + 
				classes[wordInd] + " ";
				wordInd++;
			}
			out.write(outLine.trim() + "\n");
			outLine = "";
		}
		out.close();
	}

	/**
	 * Returns the bag-of-words
	 */
	public int[] getWords(){
		return words;
	}

	public int getNumTypes(){
		return wordTypes2Int.size();
	}

	/**
	 * Returns the gold-standard tags (if they exist)
	 */
	public int[] getGoldTags(){
		int wordInd = 0;
		for (int sentence = 0; sentence < corpusGoldTags.length; sentence++){
			for (int word = 0; word < corpusGoldTags[sentence].length; word++){
				goldTags[wordInd] = corpusGoldTags[sentence][word];
				wordInd++;
			}
		}
		return goldTags;
	}

	/**
	 * Returns the features of the words in bag-of-words form 
	 */
	public Map<String, int[][]> getFeatures(){
		return featureVectors;
	}

	public String int2Word(int index){
		return int2WordTypes.get(index);
	}

	public int word2Int(String word){
		return wordTypes2Int.get(word); 
	}

	public void setMorphMethod(String morphMethod) {
		this.morphMethod = morphMethod;
	}

	public void setMorphFile(String morphFile) {
		this.morphFile = morphFile;
	}
}
